# Example 6: Customizing The Compression Process
#
# All of the previous examples have assumed the default compression parameters. This will show how to
# override some of the parameters to suit your specific needs if you so desire.

from pngpack import *
import numpy as np


# A two-dimensional array of data
RH = np.array([[60.416319  , 53.15460535, 95.85212997, 98.51983846, 25.11077670, 54.92004590, 15.33010195, 41.25880639],
       [23.37203341, 44.01536728,  8.27088314, 72.81166318, 99.56490967, 42.81624029, 98.11233113,  5.49318499],
       [32.67931869, 81.75929704, 18.05781864,  3.02975054, 24.11132572, 71.95920111, 44.2130915 , 47.64615611],
       [36.43213045, 37.73881803, 13.56776197, 72.49285515, 31.02631085, 80.51850493, 11.59999181, 66.87132086],
       [29.19554496, 56.3874622 , 54.53005075, 31.75847252, 45.45498649, 31.79411807, 93.02972296, 26.02555731],
       [75.20841979, 84.41367038, 89.24981943, 32.35982867, 11.90398185, 72.34714010, 65.33311365, 47.10862331],
       [63.87435552, 90.96749126, 91.82122211, 59.89752808, 85.04606937, 2.455975160, 96.31313360,  2.77448192],
       [43.65846042, 85.20299335, 34.87889963, 71.28725288, 95.01831718, 65.14816840, 60.50471831, 69.22919578]])


# Create a new file
myFile = Pack()


# Save the array into the new file
myFile.packGrid("humidity default", RH, units="%")
myFile.packGrid("humidity custom", RH, factor=40.95, offset=0, units="%")

# This creates some uniformity so that all relative humidity datasets will be compressed the same
# way. The default configuration will simply use the minimum and maximum values in the dataset.
# However, relative humidity can be as low as 0 % or as high as 100%. So, to ensure a strictly
# apples-to-apples comparison between multiple relative humidity datasets, it might make sense to
# enforce a predetermined set of compression factors instead of using factors that are more fluid.
#
# Note that 40.95 x 100 = 4095, which is the maximum value for three channels with the default
# separation of 16.

myFile.packGrid("humidity four-channel", RH, factor=655.35, offset=0, channels=4, units="%")

# Same thing as before, but now using 4 channels (red, green, blue, alpha; RGBA) instead of three
# channels (red, green, blue; RGB). This enables greater precision at the cost of a little more disc
# space.
#
# Again, note that 655.35 x 100 = 65535, which is the maximum value for four channels with the
# default separation of 16.
# 
# The separation can also be changed, but this is something to be very careful with. If the
# separation is too low, there is a higher risk of corruption when attempting to recover the
# original data. Either that, or the size of the file will get considerably larger.

myFile.packGrid("humidity separate", RH, separation=1, units="%")

# The best values for separation are 16 or 32, but you could probably get away with values as low as
# 8. Just make sure you know the trade off of using a separation that is too low.
#
# The only other restriction is that the separation value itself must be an integer. It's also
# recommended that the separation be an exponent of 2, but theoretically any number less than 32 will
# work.

myFile.save("RH3")

loadedFile = Pack("RH3.ppf")

RH0 = loadedFile.unpackGrid("humidity default")
RH1 = loadedFile.unpackGrid("humidity custom")
RH2 = loadedFile.unpackGrid("humidity four-channel")
RH3 = loadedFile.unpackGrid("humidity separate")

print(RH0)
print(RH1)
print(RH2)
print(RH3)

error0 = np.abs(RH0 - RH).max()
error1 = np.abs(RH1 - RH).max()
error2 = np.abs(RH2 - RH).max()
error3 = np.abs(RH3 - RH).max()

print()

print("Error for grid 0", error0)
print("Error for grid 1", error1)
print("Error for grid 2", error2)
print("Error for grid 3", error3)

# When viewing the output, notice how much more precision the four channel compression has.