################################################################################
#                                                                              #
# PNG-Pack                                                                     #
# Version 1.0.2                                                                #
#                                                                              #
# Please see "README.txt" file before using                                    #
#                                                                              #
# This Python code is provided as is. Modification of the code is permissible, #
# but you may not distribute this code without clearly crediting:              #
#                                                                              #
# David Stang                                                                  #
# (C) 2023                                                                     #
#                                                                              #
# If modifications are made, you must clearly indicate who made the            #
# modifications and what sections of the file were modified. You may not       #
# charge money for the usage of this package.                                  #
#                                                                              #
# For complete background information about this package, please read the      #
# following article:                                                           #
# http://davidstang.org/research/png.html                                      #
#                                                                              #
# Any questions regarding this package should be sent to:                      #
# david.w.stang-1@ou.edu                                                       #
#                                                                              #
################################################################################

import numpy as np
import zipfile as zipFile
import io
import os
import warnings
from PIL import Image


def generateBaseList(base, count):
	"""
	generateBaseList(base, count)

	Method for providing a list of digital place values for the provided 'base' that is 'count'
	items long. For example, generateBaseList(16, 4) would return [4096, 256, 16, 1].

	Note that 'base' must be at least 2.

	returns a list that is 'count' items long
	"""

	if (base < 2):
		raise ValueError("Number base must be 2 or greater")

	theList = [0] * count

	for c in range(0, count):
		theList[count - c - 1] = base**c

	return theList


def encode(x, factor, offset, base, separate, channels):
	"""
	encode(x, factor, offset, base, separate, channels):

	Translates a value 'x' into an encoded color based on the multiplicative factor 'factor'
	and the data offset 'offset'. The parameter 'base' must contain a list of digital place
	values. For 3 digits of hexadecimal, base will be [256, 16, 1].

	The 'separate' parameter should ideally be either 16 or 32, and this specifies how much
	spacing is between individual red, green, blue, and alpha values.

	Channels must be either 3 (for red-green-blue, RGB) or 4 (for red-green-blue-alpha RGBA). 
	The number of channels must also match the length of 'base'.

	returns a list that is either 3 or 4 entries long (depending on the value of 'channels')
	"""

	if (len(base) != channels):
		raise ValueError("Number of elements in 'base' does not match the number of channels provided: " + str(len(base)) + " != " + str(channels))

	if (channels < 3 or channels > 4):
		raise ValueError("Number of color channels must be 3 or 4")

	z = int(round(factor * (x - offset)))

	sub = [0] * channels

	for k in range(0, channels):
		if (z >= base[k]):
			value = (z // base[k])
			digit = value * separate

			if (digit > 0xf0):
				warnings.warn("Possible data loss for a provided value of " + str(x))

			sub[k] = min(digit, 0xf0)
			z = z - value * base[k]

	return sub


def decode(x, factor, offset, base, separate):
	"""
	decode(x, factor, offset, base, separate):

	Translates a value 'x' from an encoded RGB or RGBA set into a single value based on the
	multiplicative factor 'factor' and the additive value of 'offset'. Make sure that the values
	used for 'factor', 'offset', 'base', and 'separate' match those that were used to encode the
	RGB or RGBA set.

	returns a single number
	"""

	if (len(x) != len(base)):
		raise ValueError("Number of elements in 'x' does not match the number of elements in 'base': " + str(len(x)) + " != " + str(base))

	x = list(x)
	z = 0

	for i in range(0, len(x)):
		x[i] = (x[i] // separate)

		z = z + (base[i] * x[i])

	z = offset + (z / factor)

	return z


def maxFactor(bounds, channels, separation):
	"""
	maxFactor(bounds, channels, separation)

	bounds: A list or tuple that is two items long
	channels: Must be either 3 (for RGB) or 4 (for RGBA)
	separation: How much separation there is between the individual color components, should ideally be 16 or 32

	returns the calculated multiplicative factor
	"""

	if (len(bounds) < 2):
		raise ValueError("The parameter 'bounds' must contain 2 entries")
	elif (len(bounds) > 2):
		warnings.warn("Ignoring extra entries in 'bounds' parameter")

	if (channels < 3 or channels > 4):
		raise ValueError("Number of color channels must be 3 or 4")

	valueRange = abs(bounds[0] - bounds[1])

	digitBase = 1 + (0xf0 // separation)
	maxValue = digitBase**channels - 1

	if (digitBase <= 1 or maxValue <= 0):
		raise ValueError("The 'separation' parameter is too large, should use a value of 16 or 32")

	theMax = maxValue / valueRange

	return theMax


def tileGrid(data, dimensions=None):
	"""
	tileGrid(data, dimensions=None)

	Translates a 3-dimensional array 'data' into a tile set that is 2-dimensional. If specific
	dimensions are excluded, the function will determine the best dimensions to use.

	To include specific dimensions, simply do:
	tileGrid(data, dimensions=(6,6))

	returns a two-dimensional array and the tiling dimensions
	"""

	dimCount = len(data.shape)

	if (dimCount != 3):
		raise ValueError("Array must be 3-dimensional to create a tiled grid")

	tileCount = len(data[:,0,0])

	gridHeight = len(data[0,:,0])
	gridWidth = len(data[0,0,:])

	w = 0
	h = 0

	if (dimensions == None):
		bestFactor = None
		distance = None
		mid = 1 + (tileCount >> 1)

		for a in range(mid, 0, -1):
			if (tileCount % a == 0):
				d = abs(a - mid)

				if (bestFactor == None or distance == None):
					bestFactor = a
					distance = d
				elif (d < distance):
					bestFactor = a
					distance = d

		w = tileCount // bestFactor
		h = bestFactor
	else:
		if (len(dimensions) != 2):
			raise ValueError("Tiling dimension must contain two elements")

		w = dimensions[0]
		h = dimensions[1]

		if (w * h != tileCount):
			raise ValueError("Cannot tile " + str(tileCount) + " grids into a " + str(w) + " x " + str(h) + " tile set")

	tiledGrid = np.zeros([gridHeight * h, gridWidth * w])

	print(tiledGrid.shape)

	for i in range(0, tileCount):
		tileX = (i % w)
		tileY = (i // w)

		tiledGrid[tileY*gridHeight:(tileY+1)*gridHeight, tileX*gridWidth:(tileX+1)*gridWidth] = data[i,:,:]

	return tiledGrid, (w, h)


def untileGrid(data, dimensions):
	"""
	untileGrid(data, dimensions)

	Translates a tiled grid 'data' into a 3-dimensional array based on the tiling dimensions.

	For example, untileGrid(data, (2,4)) will return a 3-dimensional array that contains 8
	entries (2 x 4 = 12) in the 3rd dimension.

	returns a three-dimensional array
	"""

	if (type(dimensions) is str):
		dimensions = dimensions.split(',')
		for i in range(0, len(dimensions)):
			dimensions[i] = int(dimensions[i])

	if (len(dimensions) != 2):
		raise ValueError("Tiling dimension must contain two elements")

	w = dimensions[0]
	h = dimensions[1]

	tileCount = w * h

	totalWidth = len(data[0,:])
	totalHeight = len(data[:,0])

	gridWidth = totalWidth // w
	gridHeight = totalHeight // h

	if (totalWidth % w != 0 or totalHeight % h != 0):
		raise ValueError("Tiled grid ("+str(gridWidth)+", "+str(gridHeight)+") does not evenly divide up into a tile set that is ("+str(w)+", " + str(h)+")")

	newGrid = np.zeros([tileCount, gridHeight, gridWidth])

	for i in range(0, h):
		for j in range(0, w):
			tileIndex = w*i + j

			newGrid[tileIndex,:,:] = data[gridHeight*i:gridHeight*(i+1), gridWidth*j:gridWidth*(j+1)]

	return newGrid


class Pack():
	def __init__(self, fileName=None):
		"""
		<Pack>.__init__(fileName)

		fileName: if included, will attempt to read an archive saved under 'fileName', if
		excluded, will create an entirely new file

		returns None
		"""

		self.config = {}
		self.file = {}

		if (fileName == None):
			pass
		else:
			if (not(os.path.exists(fileName))):
				raise FileNotFoundError("'" + fileName + "' could not be located")

			archive = zipFile.ZipFile(fileName, 'r')

			rawConfig = archive.read("config.txt").decode('ascii')

			lines = rawConfig.split('\n')

			for line in lines:
				buffer = line.split(':')
				name = buffer[0]
				data = buffer[1].split(';')

				print(data)

				if (data[0] == "--" or data[0] == "None"):
					data[0] = None
				else:
					data[0] = float(data[0])

				if (data[1] == "--" or data[1] == "None"):
					data[1] = None
				else:
					data[1] = float(data[1])

				data[2] = int(data[2])
				data[3] = int(data[3])

				data = tuple(data)

				self.config[name] = data
				self.file[name] = archive.read(name)


	def packGrid(self, name, array, **kwargs):
		"""
		<Pack>.packGrid(name, array, **kwargs)

		name: a unique identifier that the compressed grid will be saved under
		array: the one-dimensional or two-dimensional array of data to be compressed

		Keyword arguments:
		factor: a multiplicative factor that will override an automatically generated value
		offset: an offset that will override an automatically generated value
		channels: must be 3 (for RGB) or 4 (for RGBA)
		separation: default value is 16, ideally should be 16 or 32
		units: an optional string that can be included with the array

		returns None
		"""
		if (':' in name):
			raise ValueError("Grid name cannot contain ':'")

		argument = dict(kwargs.items())
		keyList = ["factor", "offset", "channels", "separation", "units"]

		factor = None
		offset = None
		channels = 3
		separation = 16
		units = ""

		if ("factor" in argument):
			factor = argument['factor']
		if ("offset" in argument):
			offset = argument['offset']
		if ("channels" in argument):
			channels = argument['channels']
		if ("separation" in argument):
			separation = argument['separation']
		if ("units" in argument):
			units = argument['units']

			if (';' in units):
				raise ValueError("units cannot contain semicolons, use a comma (,) instead")

		for key in argument.keys():
			if (not(key in keyList)):
				raise ValueError("Unrecognized keyword argument '" + str(key) + "'; must be 'factor', 'offset', 'channels', 'separation', or 'units'")

		masked = 0

		if (type(array) is np.ma.core.MaskedArray):
			if (not(type(array.mask) is np.bool_)):
				masked = 1

		dimCount = len(array.shape)

		if (channels < 3 or channels > 4):
			raise ValueError("Number of color channels must be 3 or 4")

		if (int(channels) != channels):
			raise ValueError("Value for channels must be an integer")

		if (int(separation) != separation):
			raise ValueError("Value for separation must be an integer")

		color = None

		if (dimCount >= 3 or dimCount == 0):
			raise ValueError("Packed array must be 1-D or 2-D")
		elif (dimCount == 2):
			color = np.zeros([len(array[:, 0]), len(array[0, :]), channels], dtype=np.uint8)
		elif (dimCount == 1):
			color = np.zeros([len(array), 1, channels], dtype=np.uint8)
		else:
			raise ValueError("Invalid array")

		digitBase = 1 + (0xf0 // separation)
		maxValue = digitBase**channels - 1

		if (digitBase <= 1 or maxValue <= 0):
			raise ValueError("The 'separation' parameter is too large, should use a value of 16 or 32")

		baseList = generateBaseList(digitBase, channels)

		arrayMax = array.max()
		arrayMin = array.min()

		if (offset == None):
			offset = arrayMin

		if (factor == None):
			if (arrayMax == arrayMin):
				factor = 1.0
			else:
				factor = maxValue / (arrayMax - arrayMin)

		if (factor * (arrayMax - offset) > maxValue):
			warnings.warn("Provided values for 'offset' and 'factor' may cause data corruption")

		self.config[name] = (factor, offset, int(separation), int(masked), units)

		if (dimCount == 2):
			for i in range(0, len(array[:, 0])):
				for j in range(0, len(array[0, :])):
					if (masked):
						if (array.mask[i, j]):
							c = [255] * channels
						else:
							c = encode(array[i, j], factor, offset, baseList, separation, channels)
					else:
						c = encode(array[i, j], factor, offset, baseList, separation, channels)

					for a in range(0, channels):
						color[i, j, a] = c[a]

		elif (dimCount == 1):
			for i in range(0, len(array)):
				if (masked > 0):
					if (array.mask[i]):
						c = [255] * channels
					else:
						c = encode(array[i], factor, offset, baseList, separation, channels)
				else:
					c = encode(array[i], factor, offset, baseList, separation, channels)

				for a in range(0, channels):
					color[i, 0, a] = c[a]
		else:
			raise ValueError("An unknown error occurred")

		channelKey = {3: "RGB", 4: "RGBA"}

		pointer = io.BytesIO()
		theImage = Image.fromarray(color).convert(channelKey[channels])
		theImage.save(pointer, format="PNG")
		imageData = pointer.getvalue()

		self.file[name] = imageData


	def unpackGrid(self, name):
		"""
		<Pack>.unpackGrid(name)

		name: the desired dataset that has been compressed into the file

		returns either a one-dimensional or two-dimensional array
		"""
		if (not(name in self.config)):
			raise ValueError("Item '" + name + "' could not be found")

		factor = self.config[name][0]
		offset = self.config[name][1]
		separation = self.config[name][2]
		masked = self.config[name][3]
		units = self.config[name][4]

		pointer = io.BytesIO(self.file[name])
		raw = Image.open(pointer)

		theData = raw.load()
		width, height = raw.size

		result = np.zeros([height, width])
		dimCount = 2
		orientation = None

		if (width == 1):
			result = np.zeros([height])
			dimCount = 1
			orientation = "H"
		elif (height == 1):
			result = np.zeros([width])
			dimCount = 1
			orientation = "W"

		if (masked > 0):
			result = np.ma.array(result, mask=list(result))

		channels = len(theData[0, 0])

		digitBase = 1 + (0xf0 // separation)
		baseList = generateBaseList(digitBase, channels)

		if (dimCount == 2):
			for y in range(0, height):
				for x in range(0, width):
					if (masked > 0 and theData[x, y][0] >= 0xf8 and theData[x, y][1] >= 0xf8 and theData[x, y][2] >= 0xf8):
						result.mask[y, x] = True
					else:
						if (masked > 0):
							result.mask[y, x] = False

						result[y, x] = decode(theData[x, y], factor, offset, baseList, separation)
		elif (dimCount == 1):
			maxIndex = height

			if (orientation == "W"):
				maxIndex = width

			for i in range(0, maxIndex):
				if (orientation == "W"):
					if (masked > 0 and theData[i, 0][0] >= 0xf8 and theData[i, 0][1] >= 0xf8 and theData[i, 0][2] >= 0xf8):
						result.mask[i] = True
					else:
						if (masked > 0):
							result.mask[i] = False

						result[i] = decode(theData[i, 0], factor, offset, baseList, separation)
				else:
					if (masked > 0 and theData[0, i][0] >= 0xf8 and theData[0, i][1] >= 0xf8 and theData[0, i][2] >= 0xf8):
						result.mask[i] = True
					else:
						if (masked > 0):
							result.mask[i] = False

						result[i] = decode(theData[0, i], factor, offset, baseList, separation)
		else:
			raise ValueError("An unexpected error occurred")

		return result


	def mergeGrid(self, otherFile, dataKey, rename=None):
		"""
		<Pack>.mergeGrid(otherFile, dataKey, rename=None)

		otherFile: Another <Pack> object that contains the desired grid
		dataKey: The variable name within 'otherFile' belonging to the desired grid
		rename: If 'None', the grid's name will be preserved. Otherwise, the new grid will be renamed to 'rename'.

		returns None
		"""
		theConfig = otherFile.config[dataKey]
		theData = otherFile.file[dataKey]

		if (rename == None):
			rename = dataKey

		if (rename in self.file or rename in self.config):
			warnings.warn("Name '" + rename + "' already exists within this Pack, contents have been overwritten.")

		self.config[rename] = theConfig
		self.file[rename] = theData


	def getKeys(self):
		"""
		<Pack>.getKeys()

		returns all identifiers within the file sorted in alphabetical order
		"""
		keyList = list(self.config.keys())
		keyList.sort()

		return keyList


	def getPrefixCount(self, prefix):
		"""
		<Pack>.getPrefixCount(prefix)

		prefix: a string that the function will search for within the list of identifiers

		returns the number of identifiers that match the prefix + a numerical index
		"""
		n = 0

		while (prefix + str(n) in self.config):
			n = n + 1

		return n


	def getUnits(self, name):
		"""
		<Pack>.getUnits(name)

		name: the identifier of interest

		returns a string whose value for 'units' match the 'name' provided
		"""
		return self.config[name][4]


	def save(self, name):
		"""
		<Pack>.save(name)

		name: the file name that all data will be saved to on the local disc

		Note that if no extension is provided, a default extension of .ppf will be used

		returns nothing
		"""
		if (not('.' in name)):
			name = name + ".ppf"

		outFile = zipFile.ZipFile(name, 'w')

		metaData = ""

		nameKey = self.config.keys()

		for key in nameKey:
			keyData = self.config[key]
			metaData = metaData + key + ":" + str(keyData[0]) + ";" + str(keyData[1]) + ";" + str(keyData[2]) + ';' + str(keyData[3]) + ';' + str(keyData[4]) + '\n'
			outFile.writestr(key, self.file[key])

		outFile.writestr("config.txt", metaData[0:-1])

		outFile.close()


if (__name__ == "__main__"):
	print("Running as a main file is not recommended. Use 'from pngpack import *' from another Python program in the same folder as this file (see included examples).")

